/*
 * Copyright (c) 1989, 1990 by the University of Washington
 *
 * For copying and distribution information, please see the file
 * <uw-copyright.h>.
 */

#include <uw-copyright.h>
#include <stdio.h>
#include <strings.h>

#include <ardp.h>
#include <pfs.h>
#include <pprot.h>
#include <pparse.h>
#include <perrno.h>

extern int	pfs_debug;

/*
 */
add_vlink(direct,lname,l,flags)
    char	*direct;	/* Virtual name for the directory	  */
    char	*lname;		/* Name of the new link                   */
    VLINK	l;		/* Link to be inserted                    */
    int		flags;		/* Options for link to add 		  */
{
    char	*dirhst;	/* Host of current directory              */
    char	*remdir;	/* Dir on remote host	                  */
    VDIR_ST	dir_st;
    VDIR	dir = &dir_st;

    RREQ	req;	 	/* Text of request to dir server          */

    int		fwdcnt = MAX_FWD_DEPTH;
    char	fwdhst[MAX_DIR_LINESIZE];
    char	fwdfnm[MAX_DIR_LINESIZE];
    OUTPUT_ST   out_st;
    OUTPUT      out = &out_st;
    INPUT_ST    in_st;
    INPUT       in = &in_st;

    int		tmp;

    vdir_init(dir);

    /* If magic number has been modified, set it back to 0 */
    if(l->f_magic_no < 0) l->f_magic_no = 0;

    /* We must first find the directory into which the link  */
    /* will be inserted                                      */

    tmp = rd_vdir(direct,0,dir,RVD_DFILE_ONLY);
    if (tmp || (dir->links == NULL)) return(DIRSRV_NOT_DIRECTORY);
    dirhst = dir->links->host;
    remdir = dir->links->hsoname;

startover:

    req = p__start_req(dirhst);
    requesttoout(req, out);
    p__add_req(req,
           "DIRECTORY ASCII %'s\nCREATE-LINK '' %c %'s %'s %s %'s %s %'s %d",
               remdir, ((flags & AVL_INVISIBLE) ? 'I' : 
                        ((flags& AVL_UNION) ? 'U' : 'L')),
            l->target, lname, l->hosttype,l->host,
           l->hsonametype, l->hsoname,l->version);
    if (l->dest_exp) p__add_req(req, " DEST-EXP %s", timetoasn(l->dest_exp));
    p__add_req(req, "\n");
    if (l->f_magic_no) p__add_req(req, "ID REMOTE %d\n", l->f_magic_no);
    out_atrs(out, l->lattrib, 0);

    tmp = ardp_send(req,dirhst,0,ARDP_WAIT_TILL_TO);

    if(pfs_debug && tmp) {
        fprintf(stderr,"Dirsend failed: %d\n",tmp);
    }

    if(req->rcvd == NOPKT) {
        ardp_rqfree(req);
        return(perrno);
    }
    /* Here we must parse reponse - While looking at each packet */
    rreqtoin(req, in);
    while(!in_eof(in)) {
        char		*line;
        char            *next_word;

        if (tmp = in_line(in, &line, &next_word)) {
            ardp_rqfree(req);
            return(tmp);
        }
        switch (*line) {

        case 'F': /* FAILURE or FORWARDED */
            /* FORWARDED */
            if(strncmp(line,"FORWARDED",9) == 0) {
                if(fwdcnt-- <= 0) {
                    ardp_rqfree(req);
                    perrno = PFS_MAX_FWD_DEPTH;
                    return(perrno);
                }
                /* parse and start over */
                tmp = sscanf(line,"FORWARDED %*s %s %*s %s %*d %*d", 
                             fwdhst,fwdfnm);
                dirhst = stcopy(fwdhst);
                remdir = stcopy(fwdfnm);

                if(tmp < 2) {
                    ardp_rqfree(req);
                    perrno = DIRSRV_BAD_FORMAT;
                    break;
                }
                ardp_rqfree(req);
                goto startover;
            }
            /* If FAILURE or anything else scan error */
            goto scanerr;

        case 'S': /* SUCCESS */
            if(strncmp(line,"SUCCESS",7) == 0) {
                ardp_rqfree(req);
                return(PSUCCESS);
            }
            goto scanerr;

        scanerr:
        default:
            if(*line && (tmp = scan_error(line))) {
                ardp_rqfree(req);
                return(tmp);
            }
            break;
        }
    }
    perrno = DIRSRV_BAD_FORMAT;
    ardp_rqfree(req);
    return(perrno);
}
