/*
 * Copyright (c) 1993 by the University of Southern California
 *
 * For copying and distribution information, please see the file
 * <usc-copyr.h>.
 */

#include <usc-copyr.h>

#include <pfs.h>
#include <pparse.h>
#include <perrno.h>

#include <sys/types.h>
#include <sys/stat.h>

/* Frees the buffer on subsequent calls. */
wholefiletoin(FILE *file, INPUT in)
{
    struct stat st_buf;
    static char *bufp = NULL;   /* initial value */
    int tmp;                    /* temp. return value from subfunctions */
    
    if(fstat(fileno(file), &st_buf)) {
        qsprintf(p_err_string, P_ERR_STRING_SZ, 
                 "wholefiletoin(): Couldn't fstat input file");
        return perrno = PFAILURE;
    }
    /* If buffer sufficiently large, don't bother calling malloc() twice. */
    if (stsize(bufp) < st_buf.st_size + 1) {
        if (bufp) stfree(bufp);
        if((bufp = stalloc(st_buf.st_size + 1)) == NULL) {
            qsprintf(p_err_string, P_ERR_STRING_SZ, "wholefiletoin(): out of \
memory");
            return perrno = PFAILURE;
        }
    }
    tmp = fread(bufp, sizeof (char), st_buf.st_size, file);
    if (tmp != st_buf.st_size) {
        qsprintf(p_err_string, P_ERR_STRING_SZ,
                 "wholefiletoin(): fread() returned %d, not the %d expected.",
                 tmp, st_buf.st_size);
        plog(NOREQ, "%s", p_err_string);
        return perrno = PFAILURE;
    }
    bufp[st_buf.st_size] = '\0';
    in->sourcetype = IO_STRING;
    in->rreq = NULL;
    in->s = bufp;
    in->file = NULL;
    in->flags = 0;
    return PSUCCESS;
}
        
                 
